from pyglet.math import Vec2, Vec3, Vec4


class Vertex:
    def __init__(
        self,
        position: Vec3 = Vec3(),
        normal: Vec3 = Vec3(0.0, 1.0, 0.0),
        tangent: Vec3 = Vec3(),
        tex_coords: Vec2 = Vec2(),
        bones_ids: tuple[int, int, int, int] = (0, 0, 0, 0),
        weights: Vec4 = Vec4(1.0)
    ):
        self.position = position
        self.normal = normal
        self.tangent = tangent
        self.tex_coords = tex_coords
        self.bones_ids = bones_ids
        self.weights = weights



from importlib.resources import files
import pyglet
from pyglet.gl import *
from pyglet.graphics import Batch, Group
from pyglet.graphics.shader import Shader, ShaderProgram
from pyglet.graphics.vertexdomain import VertexList
from pyglet.math import Vec4


from sombra_engine.graphics import WireframeGroup
from sombra_engine.models import Mesh


class Wireframe:
    def __init__(
        self,
        mesh: Mesh,
        vertex_shader: Shader = None,
        batch: Batch = None,
        group: Group = None,
        color: Vec4 = Vec4(1.0, 1.0, 1.0, 1.0)
    ):
        self.mesh = mesh
        if not vertex_shader:
            vs_src = files('sombra_engine.shaders').joinpath(
                'default.vert'
            ).read_text()
            vertex_shader = Shader(vs_src, 'vertex')
        self.vertex_shader = vertex_shader
        self.batch = batch or pyglet.graphics.get_default_batch()
        self.group = group
        fs_src = files('sombra_engine.shaders').joinpath(
            'wireframe.frag'
        ).read_text()
        fragment_shader = Shader(fs_src, 'fragment')
        self.program = ShaderProgram(vertex_shader, fragment_shader)
        self.program['color'] = color
        self.vertex_lists = self.create_vertex_lists()

    def create_vertex_lists(self) -> list[VertexList]:
        """
        This method creates a vertex list for each vertex group and returns all
        of them in a list.

        Returns:
            list[VertexList]: The list with the newly created vertex lists.
        """
        vlists = []
        # Create a list for position
        group = WireframeGroup(self.program, parent=self.group)
        for vg in self.mesh.vertex_groups.values():
            position_list = []
            for triangle in vg.triangles:
                # Only render
                for v in triangle.vertices:
                    position_list += [v.position.x, v.position.y, v.position.z]
            vl = self.program.vertex_list(
                len(vg.triangles) * 3, GL_TRIANGLES,
                batch=self.batch, group=group,
                position=('f', position_list)
            )
            vlists.append(vl)
        return vlists

    def draw(self):
        for vl in self.vertex_lists:
            vl.draw(GL_TRIANGLES)



from pyglet.graphics import Group
from pyglet.graphics.shader import ShaderProgram
from pyglet.math import Mat4


from sombra_engine.graphics import MaterialGroup
from sombra_engine.primitives import Material


MAX_BONES = 100


class SkeletalMaterialGroup(MaterialGroup):
    def __init__(
        self, material: Material, program: ShaderProgram, matrix: Mat4,
        order: int = 0, parent: Group = None
    ):
        super().__init__(
            material=material,
            program=program,
            matrix=matrix,
            order=order,
            parent=parent
        )
        self.bones_transforms = []
        for i in range(MAX_BONES):
            self.bones_transforms.append(Mat4())


    def set_state(self):
        super().set_state()
        self.program['bones_transforms'] = self.bones_transforms



from importlib.resources import files
from pyglet.graphics import Batch, Group
from pyglet.graphics.shader import Shader, ShaderProgram
import pyglet


class Gizmo:
    def __init__(
        self,
        size:float = 1.0,
        batch:Batch = None,
        group:Group = None,
    ):
        vs_src = files('sombra_engine.shaders').joinpath(
            'gizmo.vert'
        ).read_text()
        fs_src = files('sombra_engine.shaders').joinpath(
            'gizmo.frag'
        ).read_text()
        vs = Shader(vs_src, 'vertex')
        fs = Shader(fs_src, 'fragment')
        if not batch:
            batch = pyglet.graphics.get_default_batch()
        self.batch = batch
        self.program = ShaderProgram(vs, fs)
        self.mode = pyglet.gl.GL_LINES
        self.vertex_list = self.program.vertex_list(
            6, self.mode, batch=self.batch, group=group,
            position=(
                'f', (
                    0.0, 0.0, 0.0,  size, 0.0, 0.0,
                    0.0, 0.0, 0.0,  0.0, size, 0.0,
                    0.0, 0.0, 0.0,  0.0, 0.0, size
                )
            ),
            color=(
                'Bn', (
                    200, 0, 0,  200, 0, 0,
                    0, 200, 0,  0, 200, 0,
                    0, 0, 200,  0, 0, 200
                )
            )
        )

    def draw(self):
        self.batch.draw()



from sombra_engine.app import App
from sombra_engine import camera
from sombra_engine import constants
from sombra_engine import models
from sombra_engine import primitives
from sombra_engine import scene
from sombra_engine import utils



from sombra_engine.primitives import Material, Triangle


class VertexGroup:
    def __init__(
        self,
        name: str,
        triangles: list[Triangle],
        material: Material
    ):
        self.name = name
        self.triangles = triangles
        self.material = material



from dataclasses import dataclass


from sombra_engine.primitives import Transform


@dataclass
class Pose:
    bones_transforms: list[Transform]


@dataclass
class Keyframe:
    pose: Pose
    time: float


@dataclass
class Animation:
    keyframes: list[Keyframe]
    length: float



import math as _math
from pyglet.math import Mat4
import pyglet

def create_color_tex(color: tuple[int, int, int, int]) -> pyglet.image.Texture:
    color_pattern = pyglet.image.SolidColorImagePattern(color)
    img = color_pattern.create_image(16, 16)
    return img.get_texture()

def create_black_tex() -> pyglet.image.Texture:
    return create_color_tex((0, 0, 0, 255))

def create_white_tex() -> pyglet.image.Texture:
    return create_color_tex((255, 255, 255, 255))

def create_gray_tex() -> pyglet.image.Texture:
    return create_color_tex((123, 123, 123, 255))


class Quaternion(pyglet.math.Quaternion):
    @classmethod
    def from_mat4(cls, matrix: Mat4) -> 'Quaternion':

        # 00: a, 01: b, 02: c, 03: d
        # 10: e, 11: f, 12: g, 13: h
        # 20: i, 21: j, 22: k, 23: l
        # 30: m, 31: n, 32: o, 33: p

        (a, b, c, d,
         e, f, g, h,
         i, j, k, l,
         m, n, o, p) = matrix

        (m00, m01, m02, m03,
         m10, m11, m12, m13,
         m20, m21, m22, m23,
         m30, m31, m32, m33) = matrix

        tr = m00 + m11 + m22

        if tr > 0:
            s = _math.sqrt(tr + 1.0) * 2
            w = 0.25 * s
            x = (m21 - m12) / s
            y = (m02 - m20) / s
            z = (m10 - m01) / s
        elif (m00 > m11) and (m00 > m22):
            s = _math.sqrt(1.0 + m00 - m11 - m22) * 2
            w = (m21 - m12) / s
            x = 0.25 * s
            y = (m01 + m10) / s
            z = (m02 + m20) / s
        elif m11 > m22:
            s = _math.sqrt(1.0 + m11 - m00 - m22) * 2
            w = (m02 - m20) / s
            x = (m01 + m10) / s
            y = 0.25 * s
            z = (m12 + m21) / s
        else:
            s = _math.sqrt(1.0 + m22 - m00 - m11) * 2
            w = (m10 - m01) / s
            x = (m02 + m20) / s
            y = (m12 + m21) / s
            z = 0.25 * s

        return cls(w, x, y, z)



from pyglet.gl import *
from pyglet.graphics import Group
from pyglet.graphics.shader import ShaderProgram


class WireframeGroup(Group):
    def __init__(self, program: ShaderProgram, order: int = 30, parent=None):
        super().__init__(order=order, parent=parent)
        self.program = program

    def set_state(self):
        self.program.use()
        glEnable(GL_POLYGON_OFFSET_LINE)
        glPolygonMode(GL_FRONT_AND_BACK, GL_LINE)
        glPolygonOffset(0.0, -0.001)

    def unset_state(self):
        self.program.stop()
        glPolygonMode(GL_FRONT_AND_BACK, GL_FILL)
        glDisable(GL_POLYGON_OFFSET_LINE)
        # glPolygonOffset(0.0, 0.0)



from pyglet.math import Vec3


from sombra_engine.models import Mesh
from sombra_engine.scene import Light


class Scene:
    def __init__(self):
        self.lights: list[Light] = []
        self.meshes: list[Mesh] = []

    def create_light(self, position: Vec3, color: Vec3):
        light = Light(position, color)
        self.lights.append(light)

    def remove_light(self, idx: int):
        self.lights.pop(idx)

    def add_mesh(self, mesh: Mesh):
        # change the group of each mesh so that includes the light
        self.meshes.append(mesh)



from .mtl_parser import MTLParser
from sombra_engine.primitives import Material


class MTLLoader:
    @staticmethod
    def load(filename: str) -> dict[str, Material]:
        materials = {}
        mtl_parser = MTLParser()
        mtl_parser.parse(filename)
        material_id = 1
        for name, data in mtl_parser.materials.items():
            materials[name] = Material(material_id=material_id, **data)
            material_id += 1
        return materials



from pyglet.gl import *
from pyglet.graphics import Group
from pyglet.graphics.shader import ShaderProgram
from pyglet.math import Mat4
import pyglet.resource

from sombra_engine.primitives import Material
from sombra_engine import utils


class MaterialGroup(Group):
    def __init__(
        self, material: Material, program: ShaderProgram, matrix: Mat4,
        order: int = 0, parent: Group = None
    ):
        super().__init__(order, parent)
        self.program = program
        self.material = material
        self.matrix = matrix

        # Set diffuse map
        if material.diffuse_map:
            img = pyglet.image.load(material.diffuse_map)
            self.diffuse_map = img.get_texture()
        else:
            self.diffuse_map = utils.create_white_tex()
        glBindTexture(self.diffuse_map.target, self.diffuse_map.id)
        glGenerateMipmap(self.diffuse_map.target)

        # Set ambient map
        if material.ambient_map:
            img = pyglet.image.load(material.ambient_map)
            self.ambient_map = img.get_texture()
        else:
            self.ambient_map = utils.create_white_tex()
        glBindTexture(self.ambient_map.target, self.ambient_map.id)
        glGenerateMipmap(self.ambient_map.target)

        # Set specular map
        if material.specular_map:
            img = pyglet.image.load(material.specular_map)
            self.specular_map = img.get_texture()
        else:
            self.specular_map = utils.create_black_tex()
        glBindTexture(self.specular_map.target, self.specular_map.id)
        glGenerateMipmap(self.specular_map.target)

        # Set bump map
        if material.bump_map:
            img = pyglet.image.load(material.bump_map)
            self.bump_map = img.get_texture()
        else:
            self.bump_map = utils.create_black_tex()

    def set_state(self):
        self.program.use()
        glActiveTexture(GL_TEXTURE0)
        glBindTexture(self.ambient_map.target, self.ambient_map.id)
        glActiveTexture(GL_TEXTURE1)
        glBindTexture(self.diffuse_map.target, self.diffuse_map.id)
        glActiveTexture(GL_TEXTURE2)
        glBindTexture(self.specular_map.target, self.specular_map.id)
        glActiveTexture(GL_TEXTURE3)
        glBindTexture(self.bump_map.target, self.bump_map.id)
        if 'material.ambient' in self.program._uniforms:
            self.program['material.ambient'] = self.material.ambient

        if 'material.diffuse' in self.program._uniforms:
            self.program['material.diffuse'] = self.material.diffuse

        if 'material.specular' in self.program._uniforms:
            self.program['material.specular'] = self.material.specular

        if 'material.specular_exponent' in self.program._uniforms:
            self.program['material.specular_exponent'] = \
                self.material.specular_exponent

        if 'material.bump_scale' in self.program._uniforms:
            self.program['material.bump_scale'] = self.material.bump_scale

        if 'material.has_bump_map' in self.program._uniforms:
            self.program['material.has_bump_map'] = self.material.has_bump_map

        if 'material.has_specular_map' in self.program._uniforms:
            self.program['material.has_specular_map'] = \
                self.material.has_specular_map

        if 'model' in self.program._uniforms:
            self.program['model'] = self.matrix

    def unset_state(self):
        self.program.stop()

    def __hash__(self):
        return hash(
            (self.material, self.program, self.order, self.parent)
        )

    def __eq__(self, other):
        return (
            isinstance(other, MaterialGroup) and
            self.material == other.material and
            self.program == other.program and
            self.order == other.order and
            self.parent == other.parent
        )



from pyglet.math import Vec3


class Material:
    def __init__(
        self,
        material_id: int,
        name: str = "default",
        ambient: Vec3 = Vec3(1.0, 1.0, 1.0),
        diffuse: Vec3 = Vec3(1.0, 1.0, 1.0),
        specular: Vec3 = Vec3(1.0, 1.0, 1.0),
        specular_exponent: float = 0.0,
        ior: float = 1.0,
        bump_scale: float = 0.9,
        ambient_map: str = "",
        diffuse_map: str = "",
        specular_map: str = "",
        bump_map: str = ""
    ):
        self.material_id = material_id
        self.name = name
        self.ambient = ambient
        self.diffuse = diffuse
        self.specular = specular
        self.specular_exponent = specular_exponent
        self.ior = ior
        self.bump_scale = bump_scale
        self.ambient_map = ambient_map
        self.diffuse_map = diffuse_map
        self.specular_map = specular_map
        self.bump_map = bump_map
        self.has_bump_map = bump_map != ""
        self.has_specular_map = specular_map != ""

    def __hash__(self):
        return hash((self.material_id, self.name))

    def __eq__(self, other: object) -> bool:
        return (
            isinstance(other, Material) and
            self.material_id == other.material_id and
            self.name == other.name and
            self.ambient == other.ambient and
            self.diffuse == other.diffuse and
            self.specular == other.specular and
            self.specular_exponent == other.specular_exponent and
            self.ior == other.ior and
            self.ambient_map == other.ambient_map and
            self.diffuse_map == other.diffuse_map and
            self.specular_map == other.specular_map and
            self.bump_map == other.bump_map
        )



from pudu_ui import Params
from pyglet.event import EVENT_HANDLED
from pyglet.gl import (
    GL_CULL_FACE, GL_DEPTH_TEST, GL_LESS, glClearColor, glDepthFunc,
    glDisable, glEnable
)
from pyglet.graphics import Batch, Group
from pyglet.math import Mat4, Vec3
from pyglet.window import key, Window
import pyglet


from sombra_engine.camera import FPSCamera
from sombra_engine.debug import Gizmo, Stats
from sombra_engine.models import Model


class App(Window):
    def __init__(
        self,
        caption: str = "Sombra Engine",
        is_debug: bool = False
    ):
        super().__init__(caption=caption, vsync=False)
        self.is_debug = is_debug
        self.camera = FPSCamera(
            self, position=Vec3(0.0, 0.0, 15.0), pitch=90, yaw=-90
        )
        self.batch = Batch()
        self.debug_group = Group()
        self.debug_group.visible = is_debug

        self.gizmo = Gizmo(size=10.0)
        self.debug_ui_batch = Batch()
        self.ui_projection = Mat4.orthogonal_projection(
            0.0, self.width, 0.0, self.height,
            0.0, 1000.0
        )
        stats_params = Params(x=10.0, y=10.0)
        self.stats = Stats(
            self, stats_params, self.debug_ui_batch, self.debug_group
        )

        self.models: list[Model] = []
        self.tri_count = 0

    def add_model(self, model: Model):
        self.models.append(model)
        self.tri_count += model.tri_count
        self.stats.set_tri_counts(self.tri_count)

    def draw_2d_debug_ui(self):
        temp_proj = self.projection
        temp_view = self.view
        self.projection = self.ui_projection
        self.view = Mat4()

        # Needs to disable DEPTH TEST for 2D UI
        glDisable(GL_DEPTH_TEST)
        self.stats.draw()

        self.projection = temp_proj
        self.view = temp_view

    def on_draw(self):
        glClearColor(0.0, 0.0, 0.0, 1.0)
        self.clear()

        # Draw gizmo first
        glDisable(GL_DEPTH_TEST)
        self.gizmo.draw()

        glEnable(GL_CULL_FACE)
        glEnable(GL_DEPTH_TEST)
        glDepthFunc(GL_LESS)

        self.batch.draw()
        if self.is_debug:
            # Use 2D UI here
            self.draw_2d_debug_ui()
        return EVENT_HANDLED

    def on_key_press(self, symbol, mod):
        handled = super().on_key_press(symbol, mod)
        if not handled:
            if mod & key.MOD_SHIFT and symbol == key.P:
                pyglet.image.get_buffer_manager().get_color_buffer().save(
                    'docs/screenshot.png'
                )
                handled = EVENT_HANDLED
        return handled

    def update(self, dt: float):
        self.stats.update(dt)

        # Update models
        for model in self.models:
            model.update(dt)

    def run(self, interval: float = 1.0 / 60.0):
        if not interval:
            pyglet.clock.schedule(self.update)
        else:
            pyglet.clock.schedule_interval(self.update, interval)
        pyglet.app.run(interval)



from pyglet.graphics import Batch, Group
from pyglet.graphics.shader import ShaderProgram


from sombra_engine.models import Mesh, Model, SkeletalMesh
from .obj_parser import OBJParser
from sombra_engine.primitives import VertexGroup


class OBJLoader:
    @staticmethod
    def load(
        filename: str,
        name: str = "unnamed object",
        program: ShaderProgram = None,
        scale: float = 1.0,
        batch: Batch = None,
        group: Group = None
    ) -> Model:
        obj_parser = OBJParser()
        # Parse the file
        obj_parser.parse(filename, scale=scale)

        # Now create the Model
        meshes = []
        for mesh_data in obj_parser.meshes_data.values():
            # Create Vertex Groups
            vertex_groups: dict[str, VertexGroup] = {}
            for vg_data in mesh_data['vertex_groups'].values():
                vertex_groups[vg_data['name']] = VertexGroup(
                    vg_data['name'], vg_data['triangles'], vg_data['material']
                )
            # Create Mesh
            new_mesh = Mesh(
            # new_mesh = SkeletalMesh(
                mesh_data['name'],
                vertex_groups=vertex_groups,
                materials=obj_parser.materials,
                batch=batch,
                group=group,
                program=program
            )
            meshes.append(new_mesh)
        model = Model(name, meshes)
        return model



from sombra_engine.models.mesh import Mesh
from sombra_engine.models.skeletal_mesh import SkeletalMesh
from sombra_engine.models.model import Model
from sombra_engine.models.wireframe import Wireframe
# IMPORTANT DON'T MOVE THE IMPORT OF OBJ, IT NEEDS THIS ORDER BECAUSE IT USES
# MESH AND MODEL
from sombra_engine.models import obj



from pyglet.math import Mat4


from sombra_engine.primitives import Transform


class SceneObject:
    def __init__(self, transform: Transform = Transform()):
        self.transform = transform

    def get_matrix(self) -> Mat4:
        return self.transform.get_matrix()



from sombra_engine.primitives import *
from .mesh import Mesh


class Model:
    def __init__(
        self, name: str, meshes: list[Mesh], transform: Transform = Transform(),
        parent: SceneObject = None
    ):
        self.name = name
        self.meshes = meshes
        self.transform = transform
        self.parent = parent
        self.tri_count: int = 0
        self.calculate_tri_count()

    def calculate_tri_count(self):
        self.tri_count = 0
        for mesh in self.meshes:
            self.tri_count += mesh.tri_count

    def draw(self):
        for mesh in self.meshes:
            mesh.draw()

    def update(self, dt: float):
        for mesh in self.meshes:
            mesh.update(dt)



from enum import Enum, auto


class CameraActions(Enum):
    MOVE_FORWARD = auto()
    MOVE_BACKWARDS = auto()
    MOVE_LEFT = auto()
    MOVE_RIGHT = auto()
    TURN_LEFT = auto()
    TURN_RIGHT = auto()
    TURN_UP = auto()
    TURN_DOWN = auto()



from math import cos, sin, radians
import pyglet
from pyglet.math import Mat4, Vec3, clamp
from pyglet.window import key
import weakref


SENSITIVITY = 0.3


class FPSCamera:
    def __init__(
        self, window,
        position=Vec3(0.0, 2.0, -2.0),
        target=Vec3(0.0, 0.0, 0.0),
        up=Vec3(0.0, 1.0, 0.0),
        pitch=90.0,
        yaw=90.0
    ):
        self.position = position
        self.target = target
        self.up = up

        self.speed = 10.0

        self.pitch = pitch
        self.yaw = yaw

        self.input_map = {
            key.W: "forward",
            key.S: "backward",
            key.A: "left",
            key.D: "right",
        }

        self.forward = False
        self.backward = False
        self.left = False
        self.right = False

        self._window = weakref.proxy(window)
        self._window.view = Mat4.look_at(position, target, up)
        self._window.push_handlers(self)

    def on_resize(self, width, height):
        self._window.viewport = (0, 0, *self._window.get_framebuffer_size())
        self._window.projection = Mat4.perspective_projection(
            self._window.aspect_ratio, z_near=0.1, z_far=1000, fov=45
        )
        return pyglet.event.EVENT_HANDLED

    def on_refresh(self, dt):
        # Movement
        speed = self.speed * dt
        if self.forward:
            self.position += (self.target * speed)
        if self.backward:
            self.position -= (self.target * speed)
        if self.left:
            self.position -= (self.target.cross(self.up).normalize() * speed)
        if self.right:
            self.position += (self.target.cross(self.up).normalize() * speed)

        # Look

        phi = radians(self.yaw)
        theta = radians(self.pitch)
        self.target = Vec3(
            sin(theta) * cos(phi),
            cos(theta),
            sin(theta) * sin(phi)
        )

        eye = Vec3(*self.position)
        center = Vec3(*(self.position + self.target))
        up = Vec3(*self.up)

        self._window.view = Mat4.look_at(eye, center, up)

    # Mouse input

    def on_mouse_motion(self, x, y, dx, dy):
        pass

    def on_mouse_drag(self, x, y, dx, dy, buttons, mod):
        self.yaw += dx * SENSITIVITY
        self.pitch = clamp(self.pitch - dy * SENSITIVITY, 0, 189.0)

    # Keyboard input

    def on_key_press(self, symbol, mod):
        if symbol in self.input_map:
            setattr(self, self.input_map[symbol], True)
        elif symbol == key.P:
            print(f"position: {self.position}")
            print(f"pitch: {self.pitch}, yaw: {self.yaw}")

    def on_key_release(self, symbol, mod):
        if symbol in self.input_map:
            setattr(self, self.input_map[symbol], False)



import os
import psutil
from pudu_ui import Label, LabelParams, Params, Widget
from pudu_ui.styles.fonts import FontStyle
import pudu_ui
from pyglet.graphics import Batch, Group
from pyglet.window import FPSDisplay, Window


DEBUG_FONT_SIZE = 14
TIME_TO_UPDATE_DEBUG = 0.5
LABEL_Y_MARGIN = 22


class Stats(Widget):
    def __init__(
        self, window: Window, params: Params, batch: Batch, group: Group
    ):
        super().__init__(params=params, batch=batch, group=group)
        self.fps_display = FPSDisplay(window, color=(0, 127, 0, 127))
        self.fps_label = self.fps_display.label
        self.fps_label.position = (self.x, self.y, 0.0)
        self.fps_label.font_size = DEBUG_FONT_SIZE
        debug_label_style = FontStyle(
            font_size=DEBUG_FONT_SIZE,
            weight=pudu_ui.styles.fonts.Weight.BOLD,
            color=pudu_ui.colors.Color(0, 127, 0),
            opacity=127
        )

        label_params = LabelParams(
            x=0.0,
            y=LABEL_Y_MARGIN,
            text="hola",
            style=debug_label_style
        )
        self.memory_label = Label(
            params=label_params,
            batch=batch,
            group=group,
            parent=self
        )
        label_params.y += LABEL_Y_MARGIN
        self.tri_count_label = Label(
            params=label_params,
            batch=batch,
            group=group,
            parent=self
        )

        # Add children widgets
        self.children.append(self.memory_label)
        self.children.append(self.tri_count_label)

        self.batch = batch
        self.time_to_update_debug = TIME_TO_UPDATE_DEBUG
        self.process = psutil.Process(os.getpid())

    def draw(self):
        self.fps_display.draw()
        self.batch.draw()

    def update(self, dt: float):
        self.time_to_update_debug -= dt
        if self.time_to_update_debug <= 0:
            self.time_to_update_debug = TIME_TO_UPDATE_DEBUG
            ram_used = self.process.memory_info().rss / (1024 * 1024)  # in MB
            self.memory_label.text = f"{round(ram_used, 2)} MB"
            self.invalidate()

        super().update(dt)

    def set_tri_counts(self, tri_count: int):
        self.tri_count_label.text = f"{tri_count:,} triangles".replace(
            ",", "."
        )
        self.invalidate()



from importlib.resources import files
from pyglet.gl import GL_TRIANGLES
from pyglet.graphics import Batch, Group
from pyglet.graphics.shader import Shader, ShaderProgram
from pyglet.graphics.vertexdomain import VertexList
from pyglet.math import Mat4


from sombra_engine.animations import Animation
from sombra_engine.graphics import SkeletalMaterialGroup
from sombra_engine.models import Mesh
from sombra_engine.primitives import (
    Material, SceneObject, Transform, VertexGroup
)


class Bone:
    def __init__(self, idx: int, name: str, local_bind_transform: Mat4):
        self.idx = idx
        self.name = name
        self.local_bind_transform = local_bind_transform
        self.inverse_bind_transform = None
        self.children: list[Bone] = []


class SkeletalMesh(Mesh):
    def __init__(
        self,
        name: str,
        vertex_groups: dict[str, VertexGroup] = None,
        materials: dict[str, Material] = None,
        root_bone: Bone = None,
        mode: int = GL_TRIANGLES,
        batch: Batch = None,
        group: Group = None,
        program: ShaderProgram = None,
        transform: Transform = Transform(),
        parent: SceneObject = None
    ):
        if not program:
            vs_src = files('sombra_engine.shaders').joinpath(
                'skeletal.vert'
            ).read_text()
            vert_shader = Shader(vs_src, 'vertex')

            fs_src = files('sombra_engine.shaders').joinpath(
                'blinn.frag'
            ).read_text()
            frag_shader = Shader(fs_src, 'fragment')

            program = ShaderProgram(vert_shader, frag_shader)
        super().__init__(
            name=name,
            vertex_groups=vertex_groups,
            materials=materials,
            mode=mode,
            batch=batch,
            group=group,
            program=program,
            transform=transform,
            parent=parent
        )
        self.root_bone = root_bone
        self.time = 0.0
        self.animation: Animation | None = None
        self.is_paused = False
        self.keyframes_count = 0
        self.keyframe_duration = 0.0

    def create_material_groups(self) -> dict[str, SkeletalMaterialGroup]:
        groups = {}
        for name, material in self.materials.items():
            new_group = SkeletalMaterialGroup(
                material, self.program, self.get_matrix(),
                order=0, parent=self.group
            )
            groups[name] = new_group
        return groups

    def create_vertex_lists(self) -> list[VertexList]:
        vlists = []

        for vg_name, vg in self.vertex_groups.items():
            (
                position_list, normal_list, tangent_list, tex_coords_list,
                bones_ids_list, weights_list
            ) = self.get_lists_for_vertex_group(vg_name)
            material_group = self.material_groups[vg.material.name]
            vl = self.program.vertex_list(
                len(vg.triangles) * 3, self.mode,
                batch=self.batch, group=material_group,
                position=('f', position_list),
                normal=('f', normal_list),
                tangent=('f', tangent_list),
                tex_coords=('f', tex_coords_list),
                bones_ids=('i', bones_ids_list),
                weights=('f', weights_list)
            )
            vlists.append(vl)
        return vlists

    def get_lists_for_vertex_group(self, vertex_group_name: str) -> tuple[
        list[float], list[float], list[float], list[float], list[int],
        list[float]
    ]:
        if vertex_group_name not in self.vertex_groups:
            raise KeyError(
                f"Couldn't find key {vertex_group_name} for vertex group in "
                f"mesh {self.name}"
            )
        position_list = []
        normal_list = []
        tangent_list = []
        tex_coords_list = []
        bones_ids_list = []
        weights_list = []

        for triangle in self.vertex_groups[vertex_group_name].triangles:
            for v in triangle.vertices:
                position_list += [v.position.x, v.position.y, v.position.z]
                normal_list += [v.normal.x, v.normal.y, v.normal.z]
                tangent_list += [v.tangent.x, v.tangent.y, v.tangent.z]
                tex_coords_list += [v.tex_coords.x, v.tex_coords.y]
                bones_ids_list += [*v.bones_ids]
                weights_list += [
                    v.weights.x, v.weights.y, v.weights.z, v.weights.w
                ]
        return (
            position_list, normal_list, tangent_list, tex_coords_list,
            bones_ids_list, weights_list
        )

    def set_bones_transforms(self, bones_transforms: list[Mat4]):
        # This is not convenient because it has a copy of all the transforms
        # for all material groups
        for mg in self.material_groups.values():
            mg.bones_transforms = bones_transforms

    def load_animation(self, animation: Animation):
        self.time = 0.0
        self.animation = animation
        self.keyframes_count = len(self.animation.keyframes)
        # Here we are using the same duration for every keyframe
        self.keyframe_duration = self.animation.length / self.keyframes_count

    def update(self, dt: float):
        if not self.animation or self.is_paused:
            return

        self.time += dt
        if self.time > self.animation.length:
            self.time = self.time % self.animation.length

        # Get the poses in between
        t = self.time % self.keyframe_duration
        bones_transforms = self.interpolate(t)
        self.set_bones_transforms(bones_transforms)

    def pause(self):
        self.is_paused = True

    def play(self):
        self.is_paused = False

    def interpolate(self, t: float) -> list[Mat4]:
        idx = int(self.time // self.keyframe_duration)
        prev_pose = self.animation.keyframes[idx].pose
        next_idx = idx + 1 if idx + 1 < self.keyframes_count else 0
        next_pose = self.animation.keyframes[next_idx].pose

        # interpolate between the two poses
        bones_transforms = []
        n = len(prev_pose.bones_transforms)
        for i in range(n):
            transform = Transform.interpolate(
                prev_pose.bones_transforms[i],
                next_pose.bones_transforms[i],
                t
            )
            bones_transforms.append(transform)
        return bones_transforms



import os
from pyglet.math import Vec3


from sombra_engine.constants import DEFAULT_MATERIAL_NAME


class MTLParser:
    def __init__(self):
        self.current_name = None
        self.materials = {}
        self.commands_map = {
            'newmtl': self.set_new_material,
            'Ka': self.set_ambient,
            'Kd': self.set_diffuse,
            'Ks': self.set_specular,
            'Ns': self.set_specular_exponent,
            'Ni': self.set_ior,
            'map_Ka': self.set_ambient_map,
            'map_Kd': self.set_diffuse_map,
            'map_Ks': self.set_specular_map,
            'map_bump': self.set_bump_map,
            'map_Bump': self.set_bump_map,
            'bump': self.set_bump_map
        }
        self.current_path = None

    def parse(self, filename: str):
        """
        Read an MTL file line by line parsing the commands and storing the
        information into data structures that are attributes of this class
        Args:
            filename: Path of an MTL file
        """
        self.current_path = os.path.dirname(filename)
        with open(filename) as file:
            for line in file:
                if not line or line[0] == '#':
                    continue
                # Remove end of line character
                if len(line) > 1 and line[-1] == '\n':
                    line = line[:-1]
                args = line.split(' ')
                # Remove empty strings
                args = [
                    arg.replace('\t', '')
                    for arg in args if arg != ''
                ]

                command = args[0]
                if command in self.commands_map:
                    self.commands_map[command](*args[1:])

    def set_new_material(self, name: str):
        self.current_name = name
        self.materials[name] = {'name': name}

    def set_color_by_key(self, key: str, r: str, g: str, b: str):
        if self.current_name is None:
            self.current_name = DEFAULT_MATERIAL_NAME
        self.materials[self.current_name][key] = Vec3(
            float(r), float(g), float(b)
        )

    def set_value_by_key(self, key: str, value: str):
        if self.current_name is None:
            self.current_name = DEFAULT_MATERIAL_NAME
        self.materials[self.current_name][key] = float(value)

    def set_map_by_key(self, key: str, filename: str):
        if self.current_name is None:
            self.current_name = DEFAULT_MATERIAL_NAME
        path = f"{self.current_path}/{filename}"
        self.materials[self.current_name][key] = path

    def set_ambient(self, r: str, g: str, b: str):
        self.set_color_by_key('ambient', r, g, b)

    def set_diffuse(self, r: str, g: str, b: str):
        self.set_color_by_key('diffuse', r, g, b)

    def set_specular(self, r: str, g: str, b: str):
        self.set_color_by_key('specular', r, g, b)

    def set_specular_exponent(self, value: str):
        self.set_value_by_key('specular_exponent', value)

    def set_ior(self, value: str):
        self.set_value_by_key('ior', value)

    def set_ambient_map(self, filename: str):
        self.set_map_by_key('ambient_map', filename)

    def set_diffuse_map(self, filename: str):
        self.set_map_by_key('diffuse_map', filename)

    def set_specular_map(self, filename: str):
        self.set_map_by_key('specular_map', filename)

    def set_bump_map(self, filename: str):
        self.set_map_by_key('bump_map', filename)



# import numpy as np
from pyglet.math import Mat4, Vec3
# import quaternion


class Transform:
    def __init__(
        self,
        translation: Vec3 = Vec3(),
        rotation: Vec3 = Vec3(),
        scale: Vec3 = Vec3(1.0, 1.0, 1.0)
    ):
        self.translation = translation
        self.rotation = rotation
        self.scale = scale

    def get_rotation_as_mat4(self) -> Mat4:
        rotation_x = Mat4.from_rotation(self.rotation.x, Vec3(1.0, 0.0, 0.0))
        rotation_y = Mat4.from_rotation(self.rotation.y, Vec3(0.0, 1.0, 0.0))
        rotation_z = Mat4.from_rotation(self.rotation.z, Vec3(0.0, 0.0, 1.0))
        rotation = rotation_x @ rotation_y @ rotation_z
        return rotation

    def get_matrix(self) -> Mat4:
        scale = Mat4.from_scale(self.scale)
        rotation = self.get_rotation_as_mat4()
        translation = Mat4.from_translation(self.translation)
        final = scale @ translation @ rotation
        return final

    # @staticmethod
    # def interpolate(a: 'Transform', b: 'Transform', t: float) -> Mat4:
    #     new_translation = a.translation.lerp(b.translation, t)
    #
    #     # Rotation
    #     q_a = Quaternion.from_mat4(a.get_rotation_as_mat4())
    #     q_b = Quaternion.from_mat4(b.get_rotation_as_mat4())
    #     q1 = np.quaternion(q_a.w, q_a.x, q_a.y, q_a.z)
    #     q2 = np.quaternion(q_b.w, q_b.x, q_b.y, q_b.z)
    #
    #     q_interpolated = quaternion.slerp_evaluate(q1, q2, t)
    #     final_q = Quaternion(
    #         q_interpolated.w,
    #         q_interpolated.x,
    #         q_interpolated.y,
    #         q_interpolated.z
    #     )
    #     rotation = final_q.to_mat4()
    #
    #     new_scale = a.scale.lerp(b.scale, t)
    #
    #     scale = Mat4.from_scale(new_scale)
    #     translation = Mat4.from_translation(new_translation)
    #     final = scale @ translation @ rotation
    #     return final



from pyglet.math import Vec3
from math import acos, asin, cos, radians, sin, sqrt
import math


SPEED: float = 30.0
SENSITIVITY: float = 0.3


class Camera:
    def __init__(
        self,
        position: Vec3 = Vec3(0.0, 0.0, -1.0),
        up: Vec3 = Vec3(0.0, 1.0, 0.0),
        target: Vec3 = Vec3(),
        speed: float = SPEED,
        sensitivity: float = SENSITIVITY
    ):
        self.position = position
        self.up = up
        self.target = target
        self.speed = speed
        self.sensitivity = sensitivity
        self.forward: bool = False
        self.backward: bool = False
        self.left: bool = False
        self.right: bool = False

        # Calculate direction from target and position
        self.direction: Vec3 = (target - position).normalize()

        # Calculate theta angle from direction
        dir_projection_xz = sqrt(self.direction.x ** 2 + self.direction.z ** 2)
        self.theta: float = 0.0
        if dir_projection_xz:
            self.theta = acos(self.direction.x / dir_projection_xz)
        if self.direction.z < 0.0:
            # If the direction of the camera is in negative z, theta can only
            # be in between the PI and 2 x PI range
            self.theta += math.pi

        # Calculate phi angle from direction
        self.phi: float = 0.0
        dir_projection_yz = sqrt(self.direction.y ** 2 + self.direction.z ** 2)
        if dir_projection_yz:
            self.phi = asin(self.direction.y / dir_projection_yz)

    def update(self, dt):
        pass



from sombra_engine.constants.actions import CameraActions as actions
from sombra_engine.scene.cameras import Camera


class Control:
    """
    The control for a camera handles input conveniently
    """
    def handle_action(self, target: Camera, action: actions):
        pass


class FPSControl(Control):
    def handle_action(self, target: Camera, action: actions):
        if action == actions.MOVE_FORWARD:
            target.position += target.forward * target.speed
        elif action == actions.MOVE_BACKWARDS:
            target.position -= target.backward * target.speed



import os
from pyglet.math import Vec2, Vec3


from sombra_engine.models.obj.mtl_loader import MTLLoader
from sombra_engine.primitives import Material, Triangle, Vertex


class OBJParser:
    """
    Class for parsing an OBJ file and storing the data in dictionaries.

    This class implements the parse method that reads an OBJ file and stores the
    information into dictionaries that are useful for another OBJLoader class.
    The OBJLoader class is the one responsible for creating a 3D model from an
    OBJ file, on the other hand this class (OBJParser) only deals with
    transforming the text into dictionaries, and loading the materials.

    Attributes:
        meshes_data: This dictionary has the name of a mesh as key and then
            another dictionary representing the data of that mesh as value.
        current_mesh_name: This string stores the name of the mesh that is
            currently being read in the OBJ file.
        current_vertex_group_name: This string stores the name of the vertex
            group that is currently being read in the OBJ file.
        materials: This dictionary has the name of a Material as key and that
            Material object as value.
    """

    def __init__(self):
        self.meshes_data: dict = {}
        self.current_mesh_name: str = "unnamed mesh"
        self.current_vertex_group_name: str = "unnamed vertex group"
        self.materials: dict[str, Material] = {}

    def parse(self, filename: str, scale: float = 1.0):
        """
        Parse a given OBJ file reading line by line and stores the model
        information into data structures that are attributes of this class.

        Args:
            filename: Path of an OBJ file
        """
        file = open(filename)
        path = os.path.dirname(filename)
        for line in file:
            if not line or line[0] == '#':
                continue
            # Remove end of line character
            if len(line) > 1 and line[-1] == '\n':
                line = line[:-1]
            args = line.split(' ')
            # Remove empty strings
            args = list(filter(lambda c: c != '', args))

            if args[0] == 'o':
                self.set_name(args[1])
            elif args[0] == 'v':
                self.set_vertex(args[1:], scale=scale)
            elif args[0] == 'vt':
                self.set_tex_coords(args[1:])
            elif args[0] == 'vn':
                self.set_normal(args[1:])
            elif args[0] == 'f':
                self.set_face(args[1:])
            elif args[0] == 'g':
                self.set_vertex_group(args[1])
            elif args[0] == 'mtllib':
                self.load_materials(os.path.join(path, args[1]))
            elif args[0] == 'usemtl':
                self.set_material(args[1])
        file.close()

    def get_current_mesh_data(self):
        if self.current_mesh_name not in self.meshes_data:
            # In this case the OBJ file doesn't define an 'o <modelName>'
            self.set_name(self.current_mesh_name)
        return self.meshes_data[self.current_mesh_name]

    def get_current_vertex_group(self):
        data = self.get_current_mesh_data()
        if self.current_vertex_group_name not in data['vertex_groups']:
            self.set_vertex_group(self.current_vertex_group_name)
        return data['vertex_groups'][self.current_vertex_group_name]

    def set_name(self, name: str):
        self.meshes_data[name] = {
            'name': name,
            'indices': [],
            'positions': [],
            'tex_coords': [],
            'normals': [],
            'vertex_groups': {}
        }
        self.current_mesh_name = name

    def set_vertex(self, args: list[str], scale: float = 1.0):
        data = self.get_current_mesh_data()
        positions = [float(x) for x in args]
        new_position = Vec3(*positions) * scale
        data['positions'].append(new_position)

    def set_face(self, args: list[str]):
        """
        Set a new face by reading the given arguments and stores it in the
        current mesh data dictionary. Will assume each element will always have
        two '/'.

        Args:
            args: the arguments of the set face command as a list,
                for ex. ['6/5/1', '7/3/2', '8/6/3']
                and this format f v1[/vt1][/vn1] v2[/vt2][/vn2] v3[/vt3][/vn3]
        """
        data = self.get_current_mesh_data()
        positions = data['positions']
        normals = data['normals']
        tex_coords = data['tex_coords']
        vg_data = self.get_current_vertex_group()
        new_vertices = []

        for vertex_str in args:
            # Split by '/' and transform the values to int
            indices_str = vertex_str.split('/')

            position_idx = int(indices_str[0]) - 1
            position = positions[position_idx]
            vertex = Vertex(position)
            if indices_str[1]:
                tex_coords_idx = int(indices_str[1]) - 1
                vertex.tex_coords = tex_coords[tex_coords_idx]
            if indices_str[2]:
                normal_idx = int(indices_str[2]) - 1
                vertex.normal = normals[normal_idx]
            new_vertices.append(vertex)
        new_triangle = Triangle(new_vertices)
        vg_data['triangles'].append(new_triangle)

    def set_tex_coords(self, args: list[str]):
        data = self.get_current_mesh_data()
        values = [float(x) for x in args]
        new_tex_coords = Vec2(*values[0:2])
        data['tex_coords'].append(new_tex_coords)

    def set_normal(self, args: list[str]):
        data = self.get_current_mesh_data()
        values = [float(x) for x in args]
        new_normal = Vec3(*values)
        data['normals'].append(new_normal)

    def set_vertex_group(self, name: str):
        data = self.get_current_mesh_data()
        self.current_vertex_group_name = name
        if name not in data['vertex_groups']:
            new_vg_data = {
                'name': name,
                'triangles': []
            }
            data['vertex_groups'][name] = new_vg_data

    def set_material(self, name: str):
        vg_data = self.get_current_vertex_group()
        if vg_data:
            if 'material' in vg_data:
                # In case of a vertex group that already has a material, create
                # a new vertex group
                self.current_vertex_group_name += "+"
                self.set_vertex_group(self.current_vertex_group_name)
                vg_data = self.get_current_vertex_group()
            vg_data['material'] = self.materials[name]

    def load_materials(self, filename: str):
        self.materials = MTLLoader.load(filename)



from importlib.resources import files
from pyglet.gl import *
from pyglet.graphics import Batch, Group
from pyglet.graphics.shader import Shader, ShaderProgram
from pyglet.graphics.vertexdomain import VertexList
from pyglet.math import Vec2, Vec3
import pyglet

from sombra_engine.graphics import MaterialGroup
from sombra_engine.primitives import (
    Material, SceneObject, Transform, Vertex, VertexGroup
)


class Mesh(SceneObject):
    def __init__(
        self,
        name: str,
        vertex_groups: dict[str, VertexGroup] = None,
        materials: dict[str, Material] = None,
        mode: int = GL_TRIANGLES,
        batch: Batch = None,
        group: Group = None,
        program: ShaderProgram = None,
        transform: Transform = Transform(),
        parent: SceneObject = None
    ):
        super().__init__(transform)
        self.name = name
        self.vertex_groups = vertex_groups
        self.materials = materials
        self.mode = mode
        self.batch = batch or pyglet.graphics.get_default_batch()
        self.group = group
        if not program:
            vs_src = files('sombra_engine.shaders').joinpath(
                'default.vert'
            ).read_text()
            vert_shader = Shader(vs_src, 'vertex')

            fs_src = files('sombra_engine.shaders').joinpath(
                'blinn_barycentric.frag'
            ).read_text()
            frag_shader = Shader(fs_src, 'fragment')

            program = ShaderProgram(vert_shader, frag_shader)
        self.program = program
        self.parent = parent

        self.material_groups = self.create_material_groups()
        self.vertex_lists = self.create_vertex_lists()
        self.tri_count: int = 0
        self.calculate_tri_count(vertex_groups)

    @staticmethod
    def check_tex_coords(a: Vertex, b: Vertex, c: Vertex):
        empty: Vec2 = Vec2()
        if (
            a.tex_coords == empty and
            b.tex_coords == empty and
            c.tex_coords == empty
        ):
            # Assign arbitrary texture coordinates when not defined
            a.tex_coords = Vec2(0.0, 1.0)
            b.tex_coords = Vec2(0.0, 0.0)
            c.tex_coords = Vec2(1.0, 0.0)

    def calculate_normals(self):
        """
        Calculate normal and tangent vectors for every vertex.
        Also define texture coordinates if they are not defined.
        """
        for vertex_group in self.vertex_groups.values():
            for triangle in vertex_group.triangles:
                a = triangle.vertices[0]
                b = triangle.vertices[1]
                c = triangle.vertices[2]
                self.check_tex_coords(a, b, c)
                # Formulas from:
                #   https://learnopengl.com/Advanced-Lighting/Normal-Mapping
                edge_1: Vec3 = b.position - a.position
                edge_2: Vec3 = c.position - a.position
                delta_uv1: Vec2 = b.tex_coords - a.tex_coords
                delta_uv2: Vec2 = c.tex_coords - a.tex_coords
                inverse_scalar = (
                    delta_uv1.x * delta_uv2.y - delta_uv2.x * delta_uv1.y
                )
                if inverse_scalar == 0:
                    continue
                f: float = 1.0 / inverse_scalar

                tangent_x: float = f * (
                    delta_uv2.y * edge_1.x - delta_uv1.y * edge_2.x
                )
                tangent_y: float = f * (
                    delta_uv2.y * edge_1.y - delta_uv1.y * edge_2.y
                )
                tangent_z: float = f * (
                    delta_uv2.y * edge_1.z - delta_uv1.y * edge_2.z
                )
                tangent = Vec3(tangent_x, tangent_y, tangent_z).normalize()

                a.tangent = tangent
                b.tangent = tangent
                c.tangent = tangent

    def calculate_tri_count(self, vertex_groups: dict[str, VertexGroup]):
        self.tri_count: int = 0
        for vg_name, vg in vertex_groups.items():
            self.tri_count += len(vg.triangles)

    # Create methods
    # -------------------------------------------------------------------------
    def create_material_groups(self) -> dict[str, MaterialGroup]:
        groups = {}
        for name, material in self.materials.items():
            new_group = MaterialGroup(
                material, self.program, self.get_matrix(),
                order=0, parent=self.group
            )
            groups[name] = new_group
        return groups

    def create_vertex_lists(self) -> list[VertexList]:
        """
        This method creates a vertex list for each vertex group using the
        Shader Program that this Mesh currently has, and returns all
        of them in a list.

        Returns:
            list[VertexList]: The list with the newly created vertex
                lists.
        """
        # first calculate normals
        self.calculate_normals()

        vlists = []

        for vg_name, vg in self.vertex_groups.items():
            (
                position_list, normal_list, tangent_list, tex_coords_list
            ) = self.get_lists_for_vertex_group(vg_name)
            material_group = self.material_groups[vg.material.name]
            vl = self.program.vertex_list(
                len(vg.triangles) * 3, self.mode,
                batch=self.batch, group=material_group,
                position=('f', position_list),
                normal=('f', normal_list),
                tangent=('f', tangent_list),
                tex_coords=('f', tex_coords_list)

            )
            vlists.append(vl)
        return vlists

    # Transform methods
    # -------------------------------------------------------------------------
    def update_matrix(self):
        matrix = self.get_matrix()
        for material_group in self.material_groups.values():
            material_group.matrix = matrix

    def rotate_x(self, angle: float):
        self.transform.rotation += Vec3(angle, 0.0, 0.0)
        self.update_matrix()

    def rotate_y(self, angle: float):
        self.transform.rotation += Vec3(0.0, angle, 0.0)
        self.update_matrix()

    def rotate_z(self, angle: float):
        self.transform.rotation += Vec3(0.0, 0.0, angle)
        self.update_matrix()

    def translate(self, vec: Vec3):
        self.transform.translation += vec
        self.update_matrix()

    def scale(self, vec: Vec3):
        self.transform.scale *= vec
        self.update_matrix()

    # -------------------------------------------------------------------------

    def draw(self):
        for vl in self.vertex_lists:
            vl.draw(self.mode)

    def get_lists_for_vertex_group(self, vertex_group_name: str) -> tuple[
        list[float], list[float], list[float], list[float]
    ]:
        if vertex_group_name not in self.vertex_groups:
            raise KeyError(
                f"Couldn't find key {vertex_group_name} for vertex group in "
                f"mesh {self.name}"
            )
        position_list = []
        normal_list = []
        tangent_list = []
        tex_coords_list = []
        for triangle in self.vertex_groups[vertex_group_name].triangles:
            for v in triangle.vertices:
                position_list += [v.position.x, v.position.y, v.position.z]
                normal_list += [v.normal.x, v.normal.y, v.normal.z]
                tangent_list += [v.tangent.x, v.tangent.y, v.tangent.z]
                tex_coords_list += [v.tex_coords.x, v.tex_coords.y]
        return position_list, normal_list, tangent_list, tex_coords_list

    def update(self, dt):
        pass
